<?php
/**
 * The MIMP_Folder:: class provides a set of methods for dealing with
 * folders, accounting for subscription, errors, etc.
 *
 * $Horde: mimp/lib/Folder.php,v 1.20 2005/06/08 03:22:24 selsky Exp $
 *
 * Copyright 2000-2005 Chuck Hagenbuch <chuck@horde.org>
 * Copyright 2000-2005 Jon Parise <jon@csh.rit.edu>
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @author  Jon Parise <jon@csh.rit.edu>
 * @package MIMP
 */
class MIMP_Folder {

    /**
     * Cached results from the exists() function.
     *
     * @var array
     */
    var $_existsResults = array();

    /**
     * Returns a reference to the global MIMP_Folder object, only
     * creating it if it doesn't already exist. This ensures that only
     * one MIMP_Folder instance is instantiated for any given session
     *
     * This method must be invoked as:
     *   $mimp_folder = &MIMP_Folder::singleton();
     *
     * @return MIMP_Folder  The MIMP_Folder instance.
     */
    function &singleton()
    {
        static $folder;

        if (!isset($folder)) {
            $folder = new MIMP_Folder();
        }

        return $folder;
    }

    /**
     * Create a new IMAP folder if it does not already exist, and
     * subcribe to it as well if requested.
     *
     * @param resource $stream    A valid c-client stream.
     * @param string $folder      The full utf encoded folder to be created.
     * @param boolean $subscribe  A boolean describing whether or not to use
     *                            folder subscriptions.
     *
     * @return boolean  Whether or not the folder was successfully created.
     */
    function create($stream, $folder, $subscribe)
    {
        global $conf, $notification;

        /* Make sure we are not trying to create a duplicate folder */
        if ($this->exists($stream, $folder)) {
            $notification->push(sprintf(_("The folder \"%s\" already exists"), MIMP::displayFolder($folder)), 'horde.warning');
            return false;
        }

        /* Attempt to create the mailbox */
        if (!imap_createmailbox($stream, MIMP::serverString() . $folder)) {
            $notification->push(sprintf(_("The folder \"%s\" was not created. This is what the server said"), MIMP::displayFolder($folder)) .
                            ': ' . imap_last_error(), 'horde.error');
            return false;
        }

        /* If the user uses SUBSCRIBE, then add to the subscribe list */
        if ($subscribe &&
            !imap_subscribe($stream, MIMP::serverString() . $folder)) {
            $notification->push(sprintf(_("The folder \"%s\" was created but you were not subscribed to it."), MIMP::displayFolder($folder)), 'horde.warning');
        } else {
            /* The folder creation has been successful. */
            $notification->push(sprintf(_("The folder \"%s\" was successfully created."), MIMP::displayFolder($folder)), 'horde.success');
        }

        /* Update the MIMP_Tree object. */
        require_once MIMP_BASE . '/lib/IMAP/Tree.php';
        $mimptree = &MIMP_Tree::singleton();
        if ($mimptree) {
            $mimptree->insert($folder);
        }

        return true;
    }

    /**
     * Find out if a specific folder exists or not.
     *
     * @param resource $stream  A valid c-client stream.
     * @param string $folder    The full utf encoded folder name to be checked.
     * @param boolean $recheck  Should we recheck the folder?
     *
     * @return boolean  Whether or not the folder exists.
     */
    function exists($stream, $folder, $recheck = false)
    {
        if ($recheck || !isset($this->_existsResults[$folder])) {
            $path = MIMP::serverString() . $folder;
            $res = imap_list($stream, '', $path);
            $this->_existsResults[$folder] = ($res ? in_array($path, $res) : false);
        }

        return $this->_existsResults[$folder];
    }

    /**
     * Lists folders.
     *
     * @param resource $stream    An open connection to a mail server.
     * @param string $server      The specification of the server to connect
     *                            to. Ex: {mail.foo.com/pop3:110}.
     * @param string $prefix      What to restrict the folder list to. For
     *                            instance, if the IMAP server makes your
     *                            entire home directory available, but you only
     *                            want the folders in ~/mail/, then this should
     *                            be 'mail/'.
     * @param boolean $sub        Should we list only subscribed folders?
     * @param array $filter       An list of mailboxes that should be left out
     *                            of the list.
     * @param array $hierarchies  An list of additional hierarchies that are
     *                            either outside prefix or not advertised by
     *                            default (#shared/ and similar with UW) to
     *                            include in the list.
     * @param boolean $dotfiles   If false, do not include folders beginning
     *                            with '.'
     * @param string $namespace   Namespace prefix prepended on folder names
     *                            (eg. "INBOX." in cyrus). We remove this for
     *                            cosmetic purposes.
     *
     * @return array  An array of folders, where each array alement is an
     *                associative array containing three values: 'val', with
     *                entire folder name after the server specification;
     *                'label', with the full-length folder name meant for
     *                display ($prefix has been stripped off, etc); and
     *                'abbrev', containing a shortened (26 characters max)
     *                label for display in situations where space is short.
     */
    function flist($stream, $server, $prefix, $sub = false, $filter = array(),
                   $hierarchies = array(), $dotfiles = true, $namespace = '')
    {
        global $conf, $notification;

        $list = array();

        /* Make sure we're requesting all folders. */
        $folders = $prefix . '*';

        /* Figure out which list command to use. */
        $listcmd = ($sub) ? 'imap_lsub' : 'imap_list';

        /* Compute values that will uniquely identify this list. */
        $signature = implode('|', array($listcmd, $folders));
        $full_signature = serialize(array($listcmd, $prefix, $filter, $hierarchies, $dotfiles, $namespace));

        if (!($mailboxes = $listcmd($stream, $server, $folders))) {
            $mailboxes = array();
        }

        if (is_array($hierarchies)) {
            foreach ($hierarchies as $val) {
                $hboxes = $listcmd($stream, $server, $val . '*');
                if (is_array($hboxes)) {
                    $mailboxes = array_merge($mailboxes, $hboxes);
                }
            }
        }

        /* Strip the server prefix from the list of mailboxes. */
        if (!empty($mailboxes)) {
            $new_mboxes = array();
            $namespace_len = strlen($namespace);
            $prefix_len = strlen($prefix);
            $pos = strpos($mailboxes[0], '}') + 1;

            foreach ($mailboxes as $val) {
                $label = $stripped = substr($val, $pos);
                if (substr($label, 0, $prefix_len) == $prefix) {
                    $label = substr($label, $prefix_len);
                }
                if (substr($label, 0, $namespace_len) == $namespace) {
                    $label = substr($label, $namespace_len);
                }
                $new_mboxes[$label] = $stripped;
            }

            $mailboxes = $new_mboxes;
        }

        if ($_SESSION['mimp']['base_protocol'] != 'pop3' && !empty($mailboxes)) {
            $parents = $seen = array();

            require_once 'Horde/IMAP/Sort.php';
            $imap_mailboxes = &new IMAP_Sort($_SESSION['mimp']['delimiter']);
            $imap_mailboxes->sortMailboxesByKey($mailboxes);

            foreach ($mailboxes as $label => $mailbox) {
                if (empty($seen[$mailbox])) {
                    $seen[$mailbox] = true;

                    if (!($decoded_mailbox = String::convertCharset($mailbox, 'UTF7-IMAP'))) {
                        $notification->push(sprintf(_("The folder \"%s\" contains illegal characters in its name. It may cause problems. Please see your system administrator."), $label), 'horde.warning');
                    }

                    /* Ignore INBOX, since we always tack it on first. */
                    if (strcmp($prefix . 'INBOX', $mailbox) != 0) {
                        $parts = explode($_SESSION['mimp']['delimiter'], $label);
                        $subfolder = array_pop($parts);
                        $parentparts = $parts;
                        $pstack = array();

                        while (($i = count($parentparts))) {
                            $parent = implode($_SESSION['mimp']['delimiter'], $parentparts);
                            if (!empty($parent) &&
                                ($parent != 'INBOX') &&
                                !isset($list[$parent]) &&
                                !isset($list[$prefix . $namespace . $parent]) &&
                                !isset($parents[$parent])) {
                                $parents[$parent] = true;
                                $plabel = str_repeat(' ', 4 * ($i - 1)) . String::convertCharset($parts[($i - 1)], 'UTF7-IMAP');
                                if (strlen($plabel) > 26) {
                                    $pabbrev = String::substr($plabel, 0, 10) . '...' . String::substr($plabel, -13, 13);
                                } else {
                                    $pabbrev = $plabel;
                                }
                                $pstack[$parent] = array('val' => '', 'label' => $plabel, 'abbrev' => $pabbrev);
                            }
                            array_pop($parentparts);
                        }

                        if (count($pstack)) {
                            $pstack = array_reverse($pstack);
                            foreach ($pstack as $pkey => $pval) {
                                $list[$pkey] = $pval;
                            }
                        }

                        $folded = str_repeat(' ', 4 * count($parts)) . String::convertCharset($subfolder, 'UTF7-IMAP');

                        if (strlen($folded) > 26) {
                            $abbrev = String::substr($folded, 0, 10) . '...' . String::substr($folded, -13, 13);
                        } else {
                            $abbrev = $folded;
                        }

                        if (($dotfiles || (strlen($label) > 0 && ($label[0] != '.' && !strstr($label, $_SESSION['mimp']['delimiter'] . '.'))))
                            && !in_array($mailbox, $filter)) {
                            $list[$mailbox] = array('val' => $mailbox, 'label' => $folded, 'abbrev' => $abbrev);
                        }
                    }
                }
            }
        }

        if (!in_array('INBOX', $filter)) {
            array_unshift($list, array('val' => 'INBOX', 'label' => _("Inbox"), 'abbrev' => _("Inbox")));
        }

        return $list;
    }

    /**
     * Returns an array of folders. This is a wrapper around the flist()
     * function which reduces the number of arguments needed if we can assume
     * that MIMP's full environment is present.
     *
     * @param array $filter  An array of mailboxes to ignore.
     * @param boolean $sub   If set, will be used to determine if we should
     *                       list only subscribed folders.
     *
     * @return array  The array of mailboxes returned by flist().
     */
    function flist_MIMP($filter = array(), $sub = null)
    {
        $sub = is_null($sub) ? $GLOBALS['prefs']->getValue('subscribe') : $sub;

        return $this->flist($_SESSION['mimp']['stream'], MIMP::serverString(), $_SESSION['mimp']['folders'], $sub, $filter, $_SESSION['mimp']['hierarchies'], $_SESSION['mimp']['dotfiles'], $_SESSION['mimp']['namespace']);
    }

}
